"""
Dispatchers
===========
"""

from __future__ import annotations

import sys
from asyncio import Future, Queue
from asyncio.tasks import FIRST_COMPLETED, ensure_future, gather, wait
from logging import getLogger
from typing import Any, AsyncIterator, Awaitable, Callable, List, Sequence, Tuple
from weakref import WeakSet

from anyio import create_task_group

from idom.utils import Ref

from .layout import Layout, LayoutEvent, LayoutUpdate


if sys.version_info >= (3, 7):  # pragma: no cover
    from contextlib import asynccontextmanager  # noqa
else:  # pragma: no cover
    from async_generator import asynccontextmanager


logger = getLogger(__name__)

SendCoroutine = Callable[[Any], Awaitable[None]]
RecvCoroutine = Callable[[], Awaitable[LayoutEvent]]


async def dispatch_single_view(
    layout: Layout,
    send: SendCoroutine,
    recv: RecvCoroutine,
) -> None:
    with layout:
        async with create_task_group() as task_group:
            task_group.start_soon(_single_outgoing_loop, layout, send)
            task_group.start_soon(_single_incoming_loop, layout, recv)


SharedViewDispatcher = Callable[[SendCoroutine, RecvCoroutine], Awaitable[None]]
_SharedViewDispatcherFuture = Callable[[SendCoroutine, RecvCoroutine], "Future[None]"]


@asynccontextmanager
async def create_shared_view_dispatcher(
    layout: Layout, run_forever: bool = False
) -> AsyncIterator[_SharedViewDispatcherFuture]:
    with layout:
        (
            dispatch_shared_view,
            model_state,
            all_update_queues,
        ) = await _make_shared_view_dispatcher(layout)

        dispatch_tasks: List[Future[None]] = []

        def dispatch_shared_view_soon(
            send: SendCoroutine, recv: RecvCoroutine
        ) -> Future[None]:
            future = ensure_future(dispatch_shared_view(send, recv))
            dispatch_tasks.append(future)
            return future

        yield dispatch_shared_view_soon

        gathered_dispatch_tasks = gather(*dispatch_tasks, return_exceptions=True)

        while True:
            (
                update_future,
                dispatchers_completed_future,
            ) = await _wait_until_first_complete(
                layout.render(),
                gathered_dispatch_tasks,
            )

            if dispatchers_completed_future.done():
                update_future.cancel()
                break
            else:
                update: LayoutUpdate = update_future.result()

            model_state.current = update.apply_to(model_state.current)
            # push updates to all dispatcher callbacks
            for queue in all_update_queues:
                queue.put_nowait(update)


def ensure_shared_view_dispatcher_future(
    layout: Layout,
) -> Tuple[Future[None], SharedViewDispatcher]:
    dispatcher_future: Future[SharedViewDispatcher] = Future()

    async def dispatch_shared_view_forever() -> None:
        with layout:
            (
                dispatch_shared_view,
                model_state,
                all_update_queues,
            ) = await _make_shared_view_dispatcher(layout)

            dispatcher_future.set_result(dispatch_shared_view)

            while True:
                update = await layout.render()
                model_state.current = update.apply_to(model_state.current)
                # push updates to all dispatcher callbacks
                for queue in all_update_queues:
                    queue.put_nowait(update)

    async def dispatch(send: SendCoroutine, recv: RecvCoroutine) -> None:
        await (await dispatcher_future)(send, recv)

    return ensure_future(dispatch_shared_view_forever()), dispatch


async def _make_shared_view_dispatcher(
    layout: Layout,
) -> Tuple[SharedViewDispatcher, Ref[Any], WeakSet[Queue[LayoutUpdate]]]:
    initial_update = await layout.render()
    model_state = Ref(initial_update.apply_to({}))

    # We push updates to queues instead of pushing directly to send() callbacks in
    # order to isolate the render loop from any errors dispatch callbacks might
    # raise.
    all_update_queues: WeakSet[Queue[LayoutUpdate]] = WeakSet()

    async def dispatch_shared_view(send: SendCoroutine, recv: RecvCoroutine) -> None:
        update_queue: Queue[LayoutUpdate] = Queue()
        async with create_task_group() as inner_task_group:
            all_update_queues.add(update_queue)
            await send(LayoutUpdate.create_from({}, model_state.current))
            inner_task_group.start_soon(_single_incoming_loop, layout, recv)
            inner_task_group.start_soon(_shared_outgoing_loop, send, update_queue)
        return None

    return dispatch_shared_view, model_state, all_update_queues


async def _single_outgoing_loop(layout: Layout, send: SendCoroutine) -> None:
    while True:
        await send(await layout.render())


async def _single_incoming_loop(layout: Layout, recv: RecvCoroutine) -> None:
    while True:
        await layout.dispatch(await recv())


async def _shared_outgoing_loop(
    send: SendCoroutine, queue: Queue[LayoutUpdate]
) -> None:
    while True:
        await send(await queue.get())


async def _wait_until_first_complete(
    *tasks: Awaitable[Any],
) -> Sequence[Future[Any]]:
    futures = [ensure_future(t) for t in tasks]
    await wait(futures, return_when=FIRST_COMPLETED)
    return futures



from logging.config import dictConfig
from typing import List

import typer

import idom
from idom.client import manage as manage_client

from .config import all_options
from .log import logging_config_defaults


main = typer.Typer()


@main.callback(invoke_without_command=True, no_args_is_help=True)
def root(
    version: bool = typer.Option(
        False,
        "--version",
        help="show the current version",
        show_default=False,
        is_eager=True,
    )
) -> None:
    """Command line interface for IDOM"""

    # reset logging config after Typer() has wrapped stdout
    dictConfig(logging_config_defaults())

    if version:
        typer.echo(idom.__version__)

    return None


@main.command()
def install(packages: List[str]) -> None:
    """Install a Javascript package from NPM into the client"""
    manage_client.build(packages, clean_build=False)
    return None


@main.command()
def restore() -> None:
    """Return to a fresh install of the client"""
    manage_client.restore()
    return None


@main.command()
def options() -> None:
    """Show available global options and their current values"""
    for opt in list(sorted(all_options(), key=lambda opt: opt.name)):
        name = typer.style(opt.name, bold=True)
        typer.echo(f"{name} = {opt.current}")



import json
import logging
import re
import shutil
from os.path import getmtime
from pathlib import Path
from typing import Dict, Set, Tuple, cast

from idom.config import IDOM_CLIENT_BUILD_DIR


logger = logging.getLogger(__name__)

HERE = Path(__file__).parent
APP_DIR = HERE / "app"
BACKUP_BUILD_DIR = APP_DIR / "build"

# the path relative to the build that contains import sources
IDOM_CLIENT_IMPORT_SOURCE_INFIX = "_snowpack/pkg"


def _run_build_dir_init_only_once() -> None:  # pragma: no cover
    """Initialize the runtime build directory - this should only be called once"""
    if not IDOM_CLIENT_BUILD_DIR.current.exists():
        logger.debug("creating new runtime build directory")
        IDOM_CLIENT_BUILD_DIR.current.parent.mkdir(parents=True, exist_ok=True)
        # populate the runtime build directory if it doesn't exist
        shutil.copytree(BACKUP_BUILD_DIR, IDOM_CLIENT_BUILD_DIR.current, symlinks=True)
    elif getmtime(BACKUP_BUILD_DIR) > getmtime(IDOM_CLIENT_BUILD_DIR.current):
        logger.debug("updating runtime build directory because it is out of date")
        # delete the existing runtime build because it's out of date
        shutil.rmtree(IDOM_CLIENT_BUILD_DIR.current)
        # replace it with the newer backup build (presumable from a fresh install)
        shutil.copytree(BACKUP_BUILD_DIR, IDOM_CLIENT_BUILD_DIR.current, symlinks=True)
    else:
        logger.debug("runtime build directory is up to date")


_run_build_dir_init_only_once()  # this is only ever called once at runtime!


def get_user_packages_file(app_dir: Path) -> Path:
    return app_dir / "packages" / "idom-app-react" / "src" / "user-packages.js"


def restore_build_dir_from_backup() -> None:
    target = IDOM_CLIENT_BUILD_DIR.current
    if target.exists():
        shutil.rmtree(target)
    shutil.copytree(BACKUP_BUILD_DIR, target, symlinks=True)


def replace_build_dir(source: Path) -> None:
    target = IDOM_CLIENT_BUILD_DIR.current
    if target.exists():
        shutil.rmtree(target)
    shutil.copytree(source, target, symlinks=True)


def get_package_name(pkg: str) -> str:
    return split_package_name_and_version(pkg)[0]


def split_package_name_and_version(pkg: str) -> Tuple[str, str]:
    at_count = pkg.count("@")
    if pkg.startswith("@"):
        if at_count == 1:
            return pkg, ""
        else:
            name, version = pkg[1:].split("@", 1)
            return ("@" + name), version
    elif at_count:
        name, version = pkg.split("@", 1)
        return name, version
    else:
        return pkg, ""


def build_dependencies() -> Dict[str, str]:
    package_json = IDOM_CLIENT_BUILD_DIR.current / "package.json"
    return cast(Dict[str, str], json.loads(package_json.read_text())["dependencies"])


_JS_MODULE_EXPORT_PATTERN = re.compile(
    r";?\s*export\s*{([0-9a-zA-Z_$\s,]*)}\s*;", re.MULTILINE
)
_JS_VAR = r"[a-zA-Z_$][0-9a-zA-Z_$]*"
_JS_MODULE_EXPORT_NAME_PATTERN = re.compile(
    fr";?\s*export\s+({_JS_VAR})\s+{_JS_VAR}\s*;", re.MULTILINE
)
_JS_MODULE_EXPORT_FUNC_PATTERN = re.compile(
    fr";?\s*export\s+function\s+({_JS_VAR})\s*\(.*?", re.MULTILINE
)


def find_js_module_exports_in_source(content: str) -> Set[str]:
    names: Set[str] = set()
    for match in _JS_MODULE_EXPORT_PATTERN.findall(content):
        for export in match.split(","):
            export_parts = export.split(" as ", 1)
            names.add(export_parts[-1].strip())
    names.update(_JS_MODULE_EXPORT_FUNC_PATTERN.findall(content))
    names.update(_JS_MODULE_EXPORT_NAME_PATTERN.findall(content))
    return names



"""
HTML Widgets
============
"""

from base64 import b64encode
from typing import Any, Callable, Dict, Optional, Union, overload

import idom
from idom.core.component import AbstractComponent, ComponentConstructor, component
from idom.core.vdom import (
    VdomDict,
    VdomDictConstructor,
    coalesce_attributes_and_children,
    make_vdom_constructor,
    vdom,
)


def image(
    format: str,
    value: Union[str, bytes] = "",
    attributes: Optional[Dict[str, Any]] = None,
) -> VdomDict:
    """Utility for constructing an image from a string or bytes

    The source value will automatically be encoded to base64
    """
    if format == "svg":
        format = "svg+xml"

    if isinstance(value, str):
        bytes_value = value.encode()
    else:
        bytes_value = value

    base64_value = b64encode(bytes_value).decode()
    src = f"data:image/{format};base64,{base64_value}"

    return {"tagName": "img", "attributes": {"src": src, **(attributes or {})}}


@component
def Input(
    callback: Callable[[str], None],
    type: str,
    value: str = "",
    attributes: Optional[Dict[str, Any]] = None,
    cast: Optional[Callable[[str], Any]] = None,
    ignore_empty: bool = True,
) -> VdomDict:
    """Utility for making an ``<input/>`` with a callback"""
    attrs = attributes or {}
    value, set_value = idom.hooks.use_state(value)

    events = idom.Events()

    @events.on("change")
    def on_change(event: Dict[str, Any]) -> None:
        value = event["value"]
        set_value(value)
        if not value and ignore_empty:
            return
        callback(value if cast is None else cast(value))

    return html.input({"type": type, "value": value, **attrs}, event_handlers=events)


class Html:
    """Utility for making basic HTML elements

    Many basic elements already have constructors, however accessing an attribute
    of any name on this object will return a constructor for an element with the
    same ``tagName``.

    All constructors return :class:`~idom.core.vdom.VdomDict`.
    """

    def __init__(self) -> None:
        # External sources
        self.link = make_vdom_constructor("link", allow_children=False)

        # Content sectioning
        self.style = make_vdom_constructor("style")
        self.address = make_vdom_constructor("address")
        self.article = make_vdom_constructor("article")
        self.aside = make_vdom_constructor("aside")
        self.footer = make_vdom_constructor("footer")
        self.h1 = make_vdom_constructor("h1")
        self.h2 = make_vdom_constructor("h2")
        self.h3 = make_vdom_constructor("h3")
        self.h4 = make_vdom_constructor("h4")
        self.h5 = make_vdom_constructor("h5")
        self.h6 = make_vdom_constructor("h6")
        self.header = make_vdom_constructor("header")
        self.hgroup = make_vdom_constructor("hgroup")
        self.nav = make_vdom_constructor("nav")
        self.section = make_vdom_constructor("section")

        # Text content
        self.blockquote = make_vdom_constructor("blockquote")
        self.dd = make_vdom_constructor("dd")
        self.div = make_vdom_constructor("div")
        self.dl = make_vdom_constructor("dl")
        self.dt = make_vdom_constructor("dt")
        self.figcaption = make_vdom_constructor("figcaption")
        self.figure = make_vdom_constructor("figure")
        self.hr = make_vdom_constructor("hr", allow_children=False)
        self.li = make_vdom_constructor("li")
        self.ol = make_vdom_constructor("ol")
        self.p = make_vdom_constructor("p")
        self.pre = make_vdom_constructor("pre")
        self.ul = make_vdom_constructor("ul")

        # Inline text semantics
        self.a = make_vdom_constructor("a")
        self.abbr = make_vdom_constructor("abbr")
        self.b = make_vdom_constructor("b")
        self.br = make_vdom_constructor("br", allow_children=False)
        self.cite = make_vdom_constructor("cite")
        self.code = make_vdom_constructor("code")
        self.data = make_vdom_constructor("data")
        self.em = make_vdom_constructor("em")
        self.i = make_vdom_constructor("i")
        self.kbd = make_vdom_constructor("kbd")
        self.mark = make_vdom_constructor("mark")
        self.q = make_vdom_constructor("q")
        self.s = make_vdom_constructor("s")
        self.samp = make_vdom_constructor("samp")
        self.small = make_vdom_constructor("small")
        self.span = make_vdom_constructor("span")
        self.strong = make_vdom_constructor("strong")
        self.sub = make_vdom_constructor("sub")
        self.sup = make_vdom_constructor("sup")
        self.time = make_vdom_constructor("time")
        self.u = make_vdom_constructor("u")
        self.var = make_vdom_constructor("var")

        # Image and video
        self.img = make_vdom_constructor("img", allow_children=False)
        self.audio = make_vdom_constructor("audio")
        self.video = make_vdom_constructor("video")
        self.source = make_vdom_constructor("source", allow_children=False)

        # Table content
        self.caption = make_vdom_constructor("caption")
        self.col = make_vdom_constructor("col")
        self.colgroup = make_vdom_constructor("colgroup")
        self.table = make_vdom_constructor("table")
        self.tbody = make_vdom_constructor("tbody")
        self.td = make_vdom_constructor("td")
        self.tfoot = make_vdom_constructor("tfoot")
        self.th = make_vdom_constructor("th")
        self.thead = make_vdom_constructor("thead")
        self.tr = make_vdom_constructor("tr")

        # Forms
        self.meter = make_vdom_constructor("meter")
        self.output = make_vdom_constructor("output")
        self.progress = make_vdom_constructor("progress")
        self.input = make_vdom_constructor("input", allow_children=False)
        self.button = make_vdom_constructor("button")
        self.label = make_vdom_constructor("label")
        self.fieldset = make_vdom_constructor("fieldset")
        self.legend = make_vdom_constructor("legend")

        # Interactive elements
        self.details = make_vdom_constructor("details")
        self.dialog = make_vdom_constructor("dialog")
        self.menu = make_vdom_constructor("menu")
        self.menuitem = make_vdom_constructor("menuitem")
        self.summary = make_vdom_constructor("summary")

    @overload
    @staticmethod
    def __call__(
        tag: ComponentConstructor, *attributes_and_children: Any
    ) -> AbstractComponent:
        ...

    @overload
    @staticmethod
    def __call__(tag: str, *attributes_and_children: Any) -> VdomDict:
        ...

    @staticmethod
    def __call__(
        tag: Union[str, ComponentConstructor],
        *attributes_and_children: Any,
    ) -> Union[VdomDict, AbstractComponent]:
        if isinstance(tag, str):
            return vdom(tag, *attributes_and_children)

        attributes, children = coalesce_attributes_and_children(attributes_and_children)

        if children:
            return tag(children=children, **attributes)
        else:
            return tag(**attributes)

    def __getattr__(self, tag: str) -> VdomDictConstructor:
        return make_vdom_constructor(tag)


html = Html()
"""Holds pre-made constructors for basic HTML elements

See :class:`Html` for more info.
"""



"""
Configuration Options
=====================

IDOM provides a series of configuration options that can be set using environment
variables or, for those which allow it, a programatic interface.
"""

import shutil
from pathlib import Path
from typing import Any, List

from appdirs import user_data_dir

import idom

from ._option import ALL_OPTIONS as _ALL_OPTIONS
from ._option import Option as _Option


def all_options() -> List[_Option[Any]]:
    """Get a list of all options"""
    return list(_ALL_OPTIONS)


IDOM_DEBUG_MODE = _Option(
    "IDOM_DEBUG_MODE",
    default=False,
    mutable=False,
    validator=lambda x: bool(int(x)),
)
"""This immutable option turns on/off debug mode

The string values ``1`` and ``0`` are mapped to ``True`` and ``False`` respectively.

When debug is on, extra validation measures are applied that negatively impact
performance but can be used to catch bugs. Additionally, the default log level for IDOM
is set to ``DEBUG``.
"""

IDOM_CLIENT_BUILD_DIR = _Option(
    "IDOM_CLIENT_BUILD_DIR",
    default=Path(user_data_dir(idom.__name__, idom.__author__)) / "client",
    validator=Path,
)
"""The location IDOM will use to store its client application

This directory **MUST** be treated as a black box. Downstream applications **MUST NOT**
assume anything about the structure of this directory see :mod:`idom.client.manage` for
a set of publically available APIs for working with the client.
"""

# TODO: remove this in 0.30.0
_DEPRECATED_BUILD_DIR = Path(__file__).parent / "client" / "build"
if _DEPRECATED_BUILD_DIR.exists():  # pragma: no cover
    shutil.rmtree(_DEPRECATED_BUILD_DIR)

IDOM_FEATURE_INDEX_AS_DEFAULT_KEY = _Option(
    "IDOM_FEATURE_INDEX_AS_DEFAULT_KEY",
    default=False,
    mutable=False,
    validator=lambda x: bool(int(x)),
)
"""Use the index of elements/components amongst their siblings as the default key.

In a future release this flag's default value will be set to true, and after that, this
flag will be removed entirely and the indices will always be the default key.

For more information on changes to this feature flag see: https://github.com/idom-team/idom/issues/351
"""



"""
Client Manager
==============
"""

import shutil
import subprocess
import sys
from logging import getLogger
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Iterable, List, Sequence, Set, Union

from idom.config import IDOM_CLIENT_BUILD_DIR

from . import _private


logger = getLogger(__name__)


def web_modules_dir() -> Path:
    """The directory containing all web modules

    .. warning::

        No assumptions should be made about the exact structure of this directory!
    """
    return IDOM_CLIENT_BUILD_DIR.current / "_snowpack" / "pkg"


def web_module_path(package_name: str, must_exist: bool = False) -> Path:
    """Get the :class:`Path` to a web module's source"""
    path = web_modules_dir().joinpath(*(package_name + ".js").split("/"))
    if must_exist and not path.exists():
        raise ValueError(
            f"Web module {package_name!r} does not exist at path {str(path)!r}"
        )
    return path


def web_module_exports(package_name: str) -> Set[str]:
    """Get a list of names this module exports"""
    web_module_path(package_name, must_exist=True)
    return _private.find_js_module_exports_in_source(
        web_module_path(package_name).read_text(encoding="utf-8")
    )


def web_module_exists(package_name: str) -> bool:
    """Whether a web module with a given name exists"""
    return web_module_path(package_name).exists()


def web_module_names() -> Set[str]:
    """Get the names of all installed web modules"""
    names = []
    web_mod_dir = web_modules_dir()
    for pth in web_mod_dir.glob("**/*.js"):
        rel_pth = pth.relative_to(web_mod_dir)
        if Path("common") in rel_pth.parents:
            continue
        module_path = str(rel_pth.as_posix())
        if module_path.endswith(".js"):
            module_path = module_path[:-3]
        names.append(module_path)
    return set(names)


def add_web_module(
    package_name: str,
    source: Union[Path, str],
) -> None:
    """Add a web module from source"""
    resolved_source = Path(source).resolve()
    if not resolved_source.exists():
        raise FileNotFoundError(f"Package source file does not exist: {str(source)!r}")
    target = web_module_path(package_name)
    if target.resolve() == resolved_source:
        return None  # already added
    target.parent.mkdir(parents=True, exist_ok=True)
    # this will raise an error if already exists
    target.symlink_to(resolved_source)


def remove_web_module(package_name: str, must_exist: bool = False) -> None:
    """Remove a web module"""
    web_module_path(package_name, must_exist).unlink()


def restore() -> None:
    _private.restore_build_dir_from_backup()


def build(
    packages: Sequence[str],
    clean_build: bool = True,
    skip_if_already_installed: bool = True,
) -> None:
    """Build the client"""
    package_specifiers_to_install = list(packages)
    del packages  # delete since we just renamed it

    packages_to_install = _parse_package_specs(package_specifiers_to_install)
    installed_packages = _private.build_dependencies()

    if clean_build:
        all_packages = packages_to_install
    else:
        if skip_if_already_installed:
            for pkg_name, pkg_ver in packages_to_install.items():
                if pkg_name not in installed_packages or (
                    pkg_ver and installed_packages[pkg_name] != pkg_ver
                ):
                    break
            else:
                logger.info(f"Already installed {package_specifiers_to_install}")
                logger.info("Build skipped ✅")
                return None
        all_packages = {**installed_packages, **packages_to_install}

    all_package_specifiers = [f"{p}@{v}" if v else p for p, v in all_packages.items()]

    with TemporaryDirectory() as tempdir:
        tempdir_path = Path(tempdir)
        temp_app_dir = tempdir_path / "app"
        temp_build_dir = temp_app_dir / "build"
        package_json_path = temp_app_dir / "package.json"

        # copy over the whole APP_DIR directory into the temp one
        shutil.copytree(_private.APP_DIR, temp_app_dir, symlinks=True)

        _write_user_packages_file(temp_app_dir, list(all_packages))

        logger.info("Installing dependencies...")
        _npm_install(all_package_specifiers, temp_app_dir)
        logger.info("Installed successfully ✅")

        logger.debug(f"package.json: {package_json_path.read_text()}")

        logger.info("Building client ...")
        _npm_run_build(temp_app_dir)
        logger.info("Client built successfully ✅")

        _private.replace_build_dir(temp_build_dir)

    not_discovered = set(all_packages).difference(web_module_names())
    if not_discovered:
        raise RuntimeError(  # pragma: no cover
            f"Successfuly installed {list(all_packages)} but "
            f"failed to discover {list(not_discovered)} post-install."
        )


if sys.platform == "win32" and sys.version_info[:2] == (3, 7):  # pragma: no cover

    def build(
        packages: Sequence[str],
        clean_build: bool = True,
        skip_if_already_installed: bool = True,
    ) -> None:
        msg = (
            "This feature is not available due to a bug in Python<3.8 on Windows - for "
            "more information see: https://bugs.python.org/issue31226"
        )
        try:
            import pytest
        except ImportError:
            raise NotImplementedError(msg)
        else:
            pytest.xfail(msg)


def _parse_package_specs(package_strings: Sequence[str]) -> Dict[str, str]:
    return {
        dep: ver
        for dep, ver in map(_private.split_package_name_and_version, package_strings)
    }


def _npm_install(packages: List[str], cwd: Path) -> None:
    _run_subprocess(["npm", "install"] + packages, cwd)


def _npm_run_build(cwd: Path) -> None:
    _run_subprocess(["npm", "run", "build"], cwd)


def _run_subprocess(args: List[str], cwd: Path) -> None:
    cmd, *args = args
    which_cmd = shutil.which(cmd)
    if which_cmd is None:
        raise RuntimeError(  # pragma: no cover
            f"Failed to run command - {cmd!r} is not installed."
        )
    try:
        subprocess.run(
            [which_cmd] + args,
            cwd=cwd,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError as error:  # pragma: no cover
        raise subprocess.SubprocessError(error.stderr.decode()) from error
    return None


def _write_user_packages_file(app_dir: Path, packages: Iterable[str]) -> None:
    _private.get_user_packages_file(app_dir).write_text(
        _USER_PACKAGES_FILE_TEMPLATE.format(
            imports=",".join(f'"{pkg}":import({pkg!r})' for pkg in packages)
        )
    )


_USER_PACKAGES_FILE_TEMPLATE = """// THIS FILE WAS GENERATED BY IDOM - DO NOT MODIFY
export default {{{imports}}};
"""



"""
Hooks
=====
"""

from __future__ import annotations

import asyncio
import weakref
from logging import getLogger
from threading import get_ident as get_thread_id
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from typing_extensions import Protocol

import idom
from idom.utils import Ref

from .component import AbstractComponent


__all__ = [
    "use_state",
    "use_effect",
    "use_reducer",
    "use_callback",
    "use_ref",
    "use_memo",
]

logger = getLogger(__name__)

_StateType = TypeVar("_StateType")


def use_state(
    initial_value: Union[_StateType, Callable[[], _StateType]],
) -> Tuple[
    _StateType, Callable[[Union[_StateType, Callable[[_StateType], _StateType]]], None]
]:
    """See the full :ref:`use_state` docs for details

    Parameters:
        initial_value:
            Defines the initial value of the state. A callable (accepting no arguments)
            can be used as a constructor function to avoid re-creating the initial value
            on each render.

    Returns:
        A tuple containing the current state and a function to update it.
    """
    current_state = _use_const(lambda: _CurrentState(initial_value))
    return current_state.value, current_state.dispatch


class _CurrentState(Generic[_StateType]):

    __slots__ = "value", "dispatch"

    def __init__(
        self,
        initial_value: Union[_StateType, Callable[[], _StateType]],
    ) -> None:
        if callable(initial_value):
            self.value = initial_value()
        else:
            self.value = initial_value

        hook = current_hook()

        def dispatch(
            new: Union[_StateType, Callable[[_StateType], _StateType]]
        ) -> None:
            if callable(new):
                next_value = new(self.value)
            else:
                next_value = new
            if next_value is not self.value:
                self.value = next_value
                hook.schedule_render()

        self.dispatch = dispatch


_EffectCleanFunc = Callable[[], None]
_SyncEffectFunc = Callable[[], Optional[_EffectCleanFunc]]
_AsyncEffectFunc = Callable[[], Awaitable[Optional[_EffectCleanFunc]]]
_EffectApplyFunc = Union[_SyncEffectFunc, _AsyncEffectFunc]


@overload
def use_effect(
    function: None = None, args: Optional[Sequence[Any]] = None
) -> Callable[[_EffectApplyFunc], None]:
    ...


@overload
def use_effect(
    function: _EffectApplyFunc, args: Optional[Sequence[Any]] = None
) -> None:
    ...


def use_effect(
    function: Optional[_EffectApplyFunc] = None,
    args: Optional[Sequence[Any]] = None,
) -> Optional[Callable[[_EffectApplyFunc], None]]:
    """See the full :ref:`use_effect` docs for details

    Parameters:
        function:
            Applies the effect and can return a clean-up function
        args:
            Dependencies for the effect. If provided the effect will only trigger when
            these args change.

    Returns:
        If not function is provided, a decorator. Otherwise ``None``.
    """
    hook = current_hook()
    memoize = use_memo(args=args)
    last_clean_callback: Ref[Optional[_EffectCleanFunc]] = use_ref(None)

    def add_effect(function: _EffectApplyFunc) -> None:

        if not asyncio.iscoroutinefunction(function):
            sync_function = cast(_SyncEffectFunc, function)
        else:

            async_function = cast(_AsyncEffectFunc, function)

            def sync_function() -> Optional[_EffectCleanFunc]:
                future = asyncio.ensure_future(async_function())

                def clean_future() -> None:
                    if not future.cancel():
                        clean = future.result()
                        if clean is not None:
                            clean()

                return clean_future

        def effect() -> None:
            if last_clean_callback.current is not None:
                last_clean_callback.current()

            clean = last_clean_callback.current = sync_function()
            if clean is not None:
                hook.add_effect("will_unmount", clean)

            return None

        return memoize(lambda: hook.add_effect("did_render", effect))

    if function is not None:
        add_effect(function)
        return None
    else:
        return add_effect


_ActionType = TypeVar("_ActionType")


def use_reducer(
    reducer: Callable[[_StateType, _ActionType], _StateType],
    initial_value: _StateType,
) -> Tuple[_StateType, Callable[[_ActionType], None]]:
    """See the full :ref:`use_reducer` docs for details

    Parameters:
        reducer:
            A function which applies an action to the current state in order to
            produce the next state.
        initial_value:
            The initial state value (same as for :func:`use_state`)

    Returns:
        A tuple containing the current state and a function to change it with an action
    """
    state, set_state = use_state(initial_value)
    return state, _use_const(lambda: _create_dispatcher(reducer, set_state))


def _create_dispatcher(
    reducer: Callable[[_StateType, _ActionType], _StateType],
    set_state: Callable[[Callable[[_StateType], _StateType]], None],
) -> Callable[[_ActionType], None]:
    def dispatch(action: _ActionType) -> None:
        set_state(lambda last_state: reducer(last_state, action))

    return dispatch


_CallbackFunc = TypeVar("_CallbackFunc", bound=Callable[..., Any])


@overload
def use_callback(
    function: None = None, args: Optional[Sequence[Any]] = None
) -> Callable[[_CallbackFunc], _CallbackFunc]:
    ...


@overload
def use_callback(
    function: _CallbackFunc, args: Optional[Sequence[Any]] = None
) -> _CallbackFunc:
    ...


def use_callback(
    function: Optional[_CallbackFunc] = None,
    args: Optional[Sequence[Any]] = (),
) -> Union[_CallbackFunc, Callable[[_CallbackFunc], _CallbackFunc]]:
    """See the full :ref:`use_callback` docs for details

    Parameters:
        function: the function whose identity will be preserved
        args: The identity the ``function`` will be udpated when these ``args`` change.

    Returns:
        The current function
    """
    memoize = use_memo(args=args)

    def setup(function: _CallbackFunc) -> _CallbackFunc:
        return memoize(lambda: function)

    if function is not None:
        return setup(function)
    else:
        return setup


class _LambdaCaller(Protocol):
    """MyPy doesn't know how to deal with TypeVars only used in function return"""

    def __call__(self, func: Callable[[], _StateType]) -> _StateType:
        ...


@overload
def use_memo(
    function: None = None, args: Optional[Sequence[Any]] = None
) -> _LambdaCaller:
    ...


@overload
def use_memo(
    function: Callable[[], _StateType], args: Optional[Sequence[Any]] = None
) -> _StateType:
    ...


def use_memo(
    function: Optional[Callable[[], _StateType]] = None,
    args: Optional[Sequence[Any]] = None,
) -> Union[_StateType, Callable[[Callable[[], _StateType]], _StateType]]:
    """See the full :ref:`use_memo` docs for details

    Parameters:
        function: The function to be memoized.
        args: The ``function`` will be recomputed when these args change.

    Returns:
        The current state
    """
    memo: _Memo[_StateType] = _use_const(_Memo)

    if memo.empty():
        # we need to initialize on the first run
        changed = True
        memo.args = () if args is None else args
    elif args is None:
        changed = True
        memo.args = ()
    elif (
        len(memo.args) != len(args)
        # if args are same length check identity for each item
        or any(current is not new for current, new in zip(memo.args, args))
    ):
        memo.args = args
        changed = True
    else:
        changed = False

    setup: Callable[[Callable[[], _StateType]], _StateType]

    if changed:

        def setup(function: Callable[[], _StateType]) -> _StateType:
            current_value = memo.value = function()
            return current_value

    else:

        def setup(function: Callable[[], _StateType]) -> _StateType:
            return memo.value

    if function is not None:
        return setup(function)
    else:
        return setup


class _Memo(Generic[_StateType]):
    """Simple object for storing memoization data"""

    __slots__ = "value", "args"

    value: _StateType
    args: Sequence[Any]

    def empty(self) -> bool:
        try:
            self.value
        except AttributeError:
            return True
        else:
            return False


def use_ref(initial_value: _StateType) -> Ref[_StateType]:
    """See the full :ref:`use_state` docs for details

    Parameters:
        initial_value: The value initially assigned to the reference.

    Returns:
        A :class:`Ref` object.
    """
    return _use_const(lambda: Ref(initial_value))


def _use_const(function: Callable[[], _StateType]) -> _StateType:
    return current_hook().use_state(function)


_current_life_cycle_hook: Dict[int, "LifeCycleHook"] = {}


def current_hook() -> "LifeCycleHook":
    """Get the current :class:`LifeCycleHook`"""
    try:
        return _current_life_cycle_hook[get_thread_id()]
    except KeyError as error:
        msg = "No life cycle hook is active. Are you rendering in a layout?"
        raise RuntimeError(msg) from error


class _EventEffects(NamedTuple):
    did_render: List[Callable[[], Any]]
    will_unmount: List[Callable[[], Any]]


class LifeCycleHook:
    """Defines the life cycle of a layout component.

    Components can request access to their own life cycle events and state, while layouts
    drive the life cycle forward by triggering events.
    """

    __slots__ = (
        "component",
        "_layout",
        "_schedule_render_later",
        "_current_state_index",
        "_state",
        "_rendered_atleast_once",
        "_is_rendering",
        "_event_effects",
        "__weakref__",
    )

    def __init__(
        self,
        layout: idom.core.layout.Layout,
        component: AbstractComponent,
    ) -> None:
        self.component = component
        self._layout = weakref.ref(layout)
        self._schedule_render_later = False
        self._is_rendering = False
        self._rendered_atleast_once = False
        self._current_state_index = 0
        self._state: Tuple[Any, ...] = ()
        self._event_effects = _EventEffects([], [])

    def schedule_render(self) -> None:
        if self._is_rendering:
            self._schedule_render_later = True
        else:
            self._schedule_render()
        return None

    def use_state(self, function: Callable[[], _StateType]) -> _StateType:
        if not self._rendered_atleast_once:
            # since we're not intialized yet we're just appending state
            result = function()
            self._state += (result,)
        else:
            # once finalized we iterate over each succesively used piece of state
            result = self._state[self._current_state_index]
        self._current_state_index += 1
        return result

    def add_effect(self, events: str, function: Callable[[], None]) -> None:
        for e in events.split():
            getattr(self._event_effects, e).append(function)

    def component_will_render(self) -> None:
        """The component is about to render"""
        self._is_rendering = True
        self._event_effects.will_unmount.clear()

    def component_did_render(self) -> None:
        """The component completed a render"""
        for effect in self._event_effects.did_render:
            try:
                effect()
            except Exception:
                msg = f"Post-render effect {effect} failed for {self.component}"
                logger.exception(msg)

        self._event_effects.did_render.clear()

        self._is_rendering = False
        if self._schedule_render_later:
            self._schedule_render()
        self._rendered_atleast_once = True
        self._current_state_index = 0

    def component_will_unmount(self) -> None:
        """The component is about to be removed from the layout"""
        for effect in self._event_effects.will_unmount:
            try:
                effect()
            except Exception:
                msg = f"Pre-unmount effect {effect} failed for {self.component}"
                logger.exception(msg)

        self._event_effects.will_unmount.clear()

    def set_current(self) -> None:
        """Set this hook as the active hook in this thread

        This method is called by a layout before entering the render method
        of this hook's associated component.
        """
        _current_life_cycle_hook[get_thread_id()] = self

    def unset_current(self) -> None:
        """Unset this hook as the active hook in this thread"""
        # this assertion should never fail - primarilly useful for debug
        assert _current_life_cycle_hook[get_thread_id()] is self
        del _current_life_cycle_hook[get_thread_id()]

    def _schedule_render(self) -> None:
        layout = self._layout()
        assert layout is not None
        layout.update(self.component)



"""
Test Tooling
============
"""

import logging
import re
from functools import wraps
from types import TracebackType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from urllib.parse import urlencode, urlunparse
from weakref import ref

from selenium.webdriver import Chrome
from selenium.webdriver.remote.webdriver import WebDriver

from idom.core.events import EventHandler
from idom.core.hooks import LifeCycleHook, current_hook
from idom.core.utils import hex_id
from idom.server.prefab import hotswap_server
from idom.server.proto import Server, ServerFactory
from idom.server.utils import find_available_port


__all__ = [
    "find_available_port",
    "create_simple_selenium_web_driver",
    "ServerMountPoint",
]


def create_simple_selenium_web_driver(
    driver_type: Type[WebDriver] = Chrome,
    driver_options: Optional[Any] = None,
    implicit_wait_timeout: float = 3.0,
    page_load_timeout: float = 3.0,
    window_size: Tuple[int, int] = (1080, 800),
) -> WebDriver:
    driver = driver_type(options=driver_options)

    driver.set_window_size(*window_size)
    driver.set_page_load_timeout(page_load_timeout)
    driver.implicitly_wait(implicit_wait_timeout)

    return driver


_Self = TypeVar("_Self", bound="ServerMountPoint[Any, Any]")
_Mount = TypeVar("_Mount")
_Server = TypeVar("_Server", bound=Server[Any])
_App = TypeVar("_App")
_Config = TypeVar("_Config")


class ServerMountPoint(Generic[_Mount, _Server]):
    """A context manager for imperatively mounting views to a render server when testing"""

    mount: _Mount
    server: _Server

    _log_handler: "_LogRecordCaptor"

    def __init__(
        self,
        server_type: Optional[ServerFactory[_App, _Config]] = None,
        host: str = "127.0.0.1",
        port: Optional[int] = None,
        server_config: Optional[_Config] = None,
        run_kwargs: Optional[Dict[str, Any]] = None,
        mount_and_server_constructor: "Callable[..., Tuple[_Mount, _Server]]" = hotswap_server,  # type: ignore
        app: Optional[_App] = None,
        **other_options: Any,
    ) -> None:
        self.host = host
        self.port = port or find_available_port(host)
        self._mount_and_server_constructor: "Callable[[], Tuple[_Mount, _Server]]" = (
            lambda: mount_and_server_constructor(
                server_type,
                self.host,
                self.port,
                server_config,
                run_kwargs,
                app,
                **other_options,
            )
        )

    @property
    def log_records(self) -> List[logging.LogRecord]:
        """A list of captured log records"""
        return self._log_handler.records

    def url(self, path: str = "", query: Optional[Any] = None) -> str:
        """Return a URL string pointing to the host and point of the server

        Args:
            path: the path to a resource on the server
            query: a dictionary or list of query parameters
        """
        return urlunparse(
            [
                "http",
                f"{self.host}:{self.port}",
                path,
                "",
                urlencode(query or ()),
                "",
            ]
        )

    def list_logged_exceptions(
        self,
        pattern: str = "",
        types: Union[Type[Any], Tuple[Type[Any], ...]] = Exception,
        log_level: int = logging.ERROR,
        del_log_records: bool = True,
    ) -> List[BaseException]:
        """Return a list of logged exception matching the given criteria

        Args:
            log_level: The level of log to check
            exclude_exc_types: Any exception types to ignore
            del_log_records: Whether to delete the log records for yielded exceptions
        """
        found: List[BaseException] = []
        compiled_pattern = re.compile(pattern)
        for index, record in enumerate(self.log_records):
            if record.levelno >= log_level and record.exc_info:
                error = record.exc_info[1]
                if (
                    error is not None
                    and isinstance(error, types)
                    and compiled_pattern.search(str(error))
                ):
                    if del_log_records:
                        del self.log_records[index - len(found)]
                    found.append(error)
        return found

    def __enter__(self: _Self) -> _Self:
        self._log_handler = _LogRecordCaptor()
        logging.getLogger().addHandler(self._log_handler)
        self.mount, self.server = self._mount_and_server_constructor()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.server.stop()
        logging.getLogger().removeHandler(self._log_handler)
        del self.mount, self.server
        logged_errors = self.list_logged_exceptions(del_log_records=False)
        if logged_errors:  # pragma: no cover
            raise logged_errors[0]
        return None


class _LogRecordCaptor(logging.NullHandler):
    def __init__(self) -> None:
        self.records: List[logging.LogRecord] = []
        super().__init__()

    def handle(self, record: logging.LogRecord) -> bool:
        self.records.append(record)
        return True


class HookCatcher:
    """Utility for capturing a LifeCycleHook from a component

    Example:
        .. code-block::

            hooks = HookCatcher(index_by_kwarg="key")

            @idom.component
            @hooks.capture
            def MyComponent(key):
                ...

            ... # render the component

            # grab the last render of where MyComponent(key='some_key')
            hooks.index["some_key"]
            # or grab the hook from the component's last render
            hooks.latest

        After the first render of ``MyComponent`` the ``HookCatcher`` will have
        captured the component's ``LifeCycleHook``.
    """

    latest: LifeCycleHook

    def __init__(self, index_by_kwarg: Optional[str] = None):
        self.index_by_kwarg = index_by_kwarg
        self.index: Dict[Any, LifeCycleHook] = {}

    def capture(self, render_function: Callable[..., Any]) -> Callable[..., Any]:
        """Decorator for capturing a ``LifeCycleHook`` on each render of a component"""

        # The render function holds a reference to `self` and, via the `LifeCycleHook`,
        # the component. Some tests check whether components are garbage collected, thus
        # we must use a `ref` here to ensure these checks pass once the catcher itself
        # has been collected.
        self_ref = ref(self)

        @wraps(render_function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            self = self_ref()
            assert self is not None, "Hook catcher has been garbage collected"

            hook = current_hook()
            if self.index_by_kwarg is not None:
                self.index[kwargs[self.index_by_kwarg]] = hook
            self.latest = hook
            return render_function(*args, **kwargs)

        return wrapper


class StaticEventHandler:
    """Utility for capturing the target of one event handler

    Example:
        .. code-block::

            static_handler = StaticEventHandler()

            @idom.component
            def MyComponent():
                state, set_state = idom.hooks.use_state(0)
                handler = static_handler.use(lambda event: set_state(state + 1))
                return idom.html.button({"onClick": handler}, "Click me!")

            # gives the target ID for onClick where from the last render of MyComponent
            static_handlers.target

        If you need to capture event handlers from different instances of a component
        the you should create multiple ``StaticEventHandler`` instances.

        .. code-block::

            static_handlers_by_key = {
                "first": StaticEventHandler(),
                "second": StaticEventHandler(),
            }

            @idom.component
            def Parent():
                return idom.html.div(Child(key="first"), Child(key="second"))

            @idom.component
            def Child(key):
                state, set_state = idom.hooks.use_state(0)
                handler = static_handlers_by_key[key].use(lambda event: set_state(state + 1))
                return idom.html.button({"onClick": handler}, "Click me!")

            # grab the individual targets for each instance above
            first_target = static_handlers_by_key["first"].target
            second_target = static_handlers_by_key["second"].target
    """

    def __init__(self) -> None:
        self._handler = EventHandler()

    @property
    def target(self) -> str:
        return hex_id(self._handler)

    def use(self, function: Callable[..., Any]) -> EventHandler:
        self._handler.clear()
        self._handler.add(function)
        return self._handler



"""
HTML Language Extension
=======================
"""

import ast
from typing import Any, List, Optional, Tuple, Union

import htm
from pyalect import Dialect, DialectError


class HtmlDialectTranspiler(Dialect, name="html"):
    """An HTML dialect transpiler for Python."""

    def __init__(self, filename: Optional[str] = None):
        self.filename: str = filename or "<dialect: html>"

    def transform_src(self, source: str) -> str:
        return source

    def transform_ast(self, node: ast.AST) -> ast.AST:
        new_node: ast.AST = HtmlDialectNodeTransformer(self.filename).visit(node)
        return new_node


class HtmlDialectNodeTransformer(ast.NodeTransformer):
    def __init__(self, filename: str):
        super().__init__()
        self.filename = filename

    def visit_Call(self, node: ast.Call) -> Optional[ast.AST]:
        if isinstance(node.func, ast.Name):
            if node.func.id == "html":
                if (
                    not node.keywords
                    and len(node.args) == 1
                    and isinstance(node.args[0], ast.JoinedStr)
                ):
                    try:
                        new_node = self._transform_string(node.args[0])
                    except htm.ParseError as error:
                        raise DialectError(str(error), self.filename, node.lineno)
                    return self.generic_visit(
                        ast.fix_missing_locations(ast.copy_location(new_node, node))
                    )
        return node

    def _transform_string(self, node: ast.JoinedStr) -> ast.Call:
        htm_strings: List[str] = []
        exp_nodes: List[ast.AST] = []
        for inner_node in node.values:
            if isinstance(inner_node, ast.Str):
                htm_strings.append(inner_node.s)
            elif isinstance(inner_node, ast.FormattedValue):
                if len(htm_strings) == len(exp_nodes):
                    htm_strings.append("")
                if inner_node.conversion != -1 or inner_node.format_spec:
                    exp_nodes.append(ast.JoinedStr([inner_node]))
                else:
                    exp_nodes.append(inner_node.value)

        call_stack = _HtmlCallStack()
        for op_type, *data in htm.htm_parse(htm_strings):
            getattr(self, f"_transform_htm_{op_type.lower()}")(
                exp_nodes, call_stack, *data
            )
        return call_stack.finish()

    def _transform_htm_open(
        self,
        exp_nodes: List[ast.AST],
        call_stack: "_HtmlCallStack",
        is_index: bool,
        tag_or_index: Union[str, int],
    ) -> None:
        if isinstance(tag_or_index, int):
            call_stack.begin_child(exp_nodes[tag_or_index])
        else:
            call_stack.begin_child(ast.Str(tag_or_index))

    def _transform_htm_close(
        self, exp_nodes: List[ast.AST], call_stack: "_HtmlCallStack"
    ) -> None:
        call_stack.end_child()

    def _transform_htm_spread(
        self, exp_nodes: List[ast.AST], call_stack: "_HtmlCallStack", _: Any, index: int
    ) -> None:
        call_stack.add_attributes(None, exp_nodes[index])

    def _transform_htm_prop_single(
        self,
        exp_nodes: List[ast.AST],
        call_stack: "_HtmlCallStack",
        attr: str,
        is_index: bool,
        value_or_index: Union[str, int],
    ) -> None:
        if isinstance(value_or_index, bool):
            const = ast.NameConstant(value_or_index)
            call_stack.add_attributes(ast.Str(attr), const)
        elif isinstance(value_or_index, int):
            call_stack.add_attributes(ast.Str(attr), exp_nodes[value_or_index])
        else:
            call_stack.add_attributes(ast.Str(attr), ast.Str(value_or_index))

    def _transform_htm_prop_multi(
        self,
        exp_nodes: List[ast.AST],
        call_stack: "_HtmlCallStack",
        attr: str,
        items: Tuple[Tuple[bool, Union[str, int]]],
    ) -> None:
        op_root = current_op = ast.BinOp(None, None, None)
        for _, value_or_index in items:
            if isinstance(value_or_index, str):
                current_op.right = ast.BinOp(ast.Str(value_or_index), ast.Add(), None)
            else:
                current_op.right = ast.BinOp(exp_nodes[value_or_index], ast.Add(), None)
            last_op = current_op
            current_op = current_op.right
        last_op.right = current_op.left
        call_stack.add_attributes(ast.Str(attr), op_root.right)

    def _transform_htm_child(
        self,
        exp_nodes: List[ast.AST],
        call_stack: "_HtmlCallStack",
        is_index: bool,
        child_or_index: Union[str, int],
    ) -> None:
        if isinstance(child_or_index, int):
            call_stack.add_child(exp_nodes[child_or_index])
        else:
            call_stack.add_child(ast.Str(child_or_index))


class _HtmlCallStack:
    def __init__(self) -> None:
        self._root = self._new(ast.Str())
        self._stack: List[ast.Call] = [self._root]

    def begin_child(self, tag: ast.AST) -> None:
        new = self._new(tag)
        last = self._stack[-1]
        children = last.args[2].elts  # type: ignore
        children.append(new)
        self._stack.append(new)

    def add_child(self, child: ast.AST) -> None:
        current = self._stack[-1]
        children = current.args[2].elts  # type: ignore
        children.append(child)

    def add_attributes(self, key: Optional[ast.Str], value: ast.AST) -> None:
        current = self._stack[-1]
        attributes: ast.Dict = current.args[1]  # type: ignore
        attributes.keys.append(key)
        attributes.values.append(value)  # type: ignore

    def end_child(self) -> None:
        self._stack.pop(-1)

    def finish(self) -> ast.Call:
        root = self._root
        self._root = self._new(ast.Str())
        self._stack.clear()
        return root.args[2].elts[0]  # type: ignore

    @staticmethod
    def _new(tag: ast.AST) -> ast.Call:
        args = [tag, ast.Dict([], []), ast.List([], ast.Load())]
        return ast.Call(ast.Name("html", ast.Load()), args, [])



from pkg_resources import DistributionNotFound as _DistributionNotFound
from pkg_resources import get_distribution as _get_distribution


try:
    __version__: str = _get_distribution(__name__).version
except _DistributionNotFound:  # pragma: no cover
    # package is not installed
    __version__ = "0.0.0"

__author__ = "idom-team"

from . import config, log
from .client.module import Import, Module, install
from .core import hooks
from .core.component import Component, component
from .core.events import Events, event
from .core.layout import Layout
from .core.vdom import VdomDict, vdom
from .server.prefab import run
from .utils import Ref, html_to_vdom
from .widgets.html import html
from .widgets.utils import hotswap, multiview


# try to automatically setup the dialect's import hook
try:
    import htm
    import pyalect
    import tagged
except ImportError:  # pragma: no cover
    pass
else:
    from . import dialect

    del pyalect
    del tagged
    del htm


__all__ = [
    "run",
    "component",
    "Component",
    "event",
    "Events",
    "dialect",
    "html",
    "hooks",
    "Layout",
    "server",
    "Ref",
    "vdom",
    "Module",
    "Import",
    "hotswap",
    "multiview",
    "html_to_vdom",
    "VdomDict",
    "widgets",
    "client",
    "install",
    "log",
    "config",
]



"""
Layout
"""

from __future__ import annotations

import abc
import asyncio
from collections import Counter
from functools import wraps
from logging import getLogger
from typing import Any, Dict, Iterator, List, NamedTuple, Optional, Set, Tuple, TypeVar
from weakref import ref

from jsonpatch import apply_patch, make_patch
from typing_extensions import TypedDict

from idom.config import IDOM_DEBUG_MODE, IDOM_FEATURE_INDEX_AS_DEFAULT_KEY

from .component import AbstractComponent
from .events import EventHandler
from .hooks import LifeCycleHook
from .utils import hex_id
from .vdom import validate_vdom


logger = getLogger(__name__)


class LayoutUpdate(NamedTuple):
    """An object describing an update to a :class:`Layout`"""

    path: str
    changes: List[Dict[str, Any]]

    def apply_to(self, model: Any) -> Any:
        """Return the model resulting from the changes in this update"""
        return apply_patch(
            model, [{**c, "path": self.path + c["path"]} for c in self.changes]
        )

    @classmethod
    def create_from(cls, source: Any, target: Any) -> "LayoutUpdate":
        return cls("", make_patch(source, target).patch)


class LayoutEvent(NamedTuple):
    target: str
    """The ID of the event handler."""
    data: List[Any]
    """A list of event data passed to the event handler."""


_Self = TypeVar("_Self", bound="Layout")


class Layout:

    __slots__ = [
        "root",
        "_event_handlers",
        "_rendering_queue",
        "_model_state_by_component_id",
    ]

    if not hasattr(abc.ABC, "__weakref__"):  # pragma: no cover
        __slots__.append("__weakref__")

    def __init__(self, root: "AbstractComponent") -> None:
        super().__init__()
        if not isinstance(root, AbstractComponent):
            raise TypeError("Expected an AbstractComponent, not %r" % root)
        self.root = root

    def __enter__(self: _Self) -> _Self:
        # create attributes here to avoid access before entering context manager
        self._event_handlers: Dict[str, EventHandler] = {}
        self._rendering_queue = _ComponentQueue()
        self._model_state_by_component_id: Dict[int, _ModelState] = {
            id(self.root): _ModelState(None, -1, "", LifeCycleHook(self, self.root))
        }
        self._rendering_queue.put(self.root)
        return self

    def __exit__(self, *exc: Any) -> None:
        root_state = self._model_state_by_component_id[id(self.root)]
        self._unmount_model_states([root_state])

        # delete attributes here to avoid access after exiting context manager
        del self._event_handlers
        del self._rendering_queue
        del self._model_state_by_component_id

        return None

    def update(self, component: "AbstractComponent") -> None:
        self._rendering_queue.put(component)
        return None

    async def dispatch(self, event: LayoutEvent) -> None:
        # It is possible for an element in the frontend to produce an event
        # associated with a backend model that has been deleted. We only handle
        # events if the element and the handler exist in the backend. Otherwise
        # we just ignore the event.
        handler = self._event_handlers.get(event.target)

        if handler is not None:
            try:
                await handler(event.data)
            except Exception:
                logger.exception(f"Failed to execute event handler {handler}")
        else:
            logger.info(
                f"Ignored event - handler {event.target!r} does not exist or its component unmounted"
            )

    async def render(self) -> LayoutUpdate:
        while True:
            component = await self._rendering_queue.get()
            if id(component) in self._model_state_by_component_id:
                return self._create_layout_update(component)
            else:
                logger.info(
                    f"Did not render component - {component} already unmounted or does not belong to this layout"
                )

    if IDOM_DEBUG_MODE.current:
        # If in debug mode inject a function that ensures all returned updates
        # contain valid VDOM models. We only do this in debug mode in order to
        # avoid unnecessarily impacting performance.

        _debug_render = render

        @wraps(_debug_render)
        async def render(self) -> LayoutUpdate:
            # Ensure that the model is valid VDOM on each render
            result = await self._debug_render()
            validate_vdom(self._model_state_by_component_id[id(self.root)].model)
            return result

    def _create_layout_update(self, component: AbstractComponent) -> LayoutUpdate:
        old_state = self._model_state_by_component_id[id(component)]
        new_state = old_state.new(None, component)

        self._render_component(old_state, new_state, component)
        changes = make_patch(getattr(old_state, "model", {}), new_state.model).patch

        # hook effects must run after the update is complete
        for state in new_state.iter_children():
            if hasattr(state, "life_cycle_hook"):
                state.life_cycle_hook.component_did_render()

        return LayoutUpdate(path=new_state.patch_path, changes=changes)

    def _render_component(
        self,
        old_state: Optional[_ModelState],
        new_state: _ModelState,
        component: AbstractComponent,
    ) -> None:
        life_cycle_hook = new_state.life_cycle_hook
        life_cycle_hook.component_will_render()
        try:
            life_cycle_hook.set_current()
            try:
                raw_model = component.render()
            finally:
                life_cycle_hook.unset_current()
            self._render_model(old_state, new_state, raw_model)
        except Exception as error:
            logger.exception(f"Failed to render {component}")
            new_state.model = {"tagName": "__error__", "children": [str(error)]}

        if old_state is not None and old_state.component is not component:
            del self._model_state_by_component_id[id(old_state.component)]
        self._model_state_by_component_id[id(component)] = new_state

        try:
            parent = new_state.parent
        except AttributeError:
            pass
        else:
            key, index = new_state.key, new_state.index
            if old_state is not None:
                assert (key, index) == (old_state.key, old_state.index,), (
                    "state mismatch during component update - "
                    f"key {key!r}!={old_state.key} "
                    f"or index {index}!={old_state.index}"
                )
            parent.children_by_key[key] = new_state
            # need to do insertion in case where old_state is None and we're appending
            parent.model["children"][index : index + 1] = [new_state.model]

    def _render_model(
        self,
        old_state: Optional[_ModelState],
        new_state: _ModelState,
        raw_model: Any,
    ) -> None:
        new_state.model = {"tagName": raw_model["tagName"]}

        self._render_model_attributes(old_state, new_state, raw_model)
        self._render_model_children(old_state, new_state, raw_model.get("children", []))

        if "key" in raw_model:
            new_state.model["key"] = raw_model["key"]
        if "importSource" in raw_model:
            new_state.model["importSource"] = raw_model["importSource"]

    def _render_model_attributes(
        self,
        old_state: Optional[_ModelState],
        new_state: _ModelState,
        raw_model: Dict[str, Any],
    ) -> None:
        # extract event handlers from 'eventHandlers' and 'attributes'
        handlers_by_event: Dict[str, EventHandler] = {}

        if "eventHandlers" in raw_model:
            handlers_by_event.update(raw_model["eventHandlers"])

        if "attributes" in raw_model:
            attrs = new_state.model["attributes"] = raw_model["attributes"].copy()
            for k, v in list(attrs.items()):
                if callable(v):
                    if not isinstance(v, EventHandler):
                        h = handlers_by_event[k] = EventHandler()
                        h.add(attrs.pop(k))
                    else:
                        h = attrs.pop(k)
                        handlers_by_event[k] = h

        if old_state is None:
            self._render_model_event_handlers_without_old_state(
                new_state, handlers_by_event
            )
            return None

        for old_event in set(old_state.targets_by_event).difference(handlers_by_event):
            old_target = old_state.targets_by_event[old_event]
            del self._event_handlers[old_target]

        if not handlers_by_event:
            return None

        model_event_handlers = new_state.model["eventHandlers"] = {}
        for event, handler in handlers_by_event.items():
            target = old_state.targets_by_event.get(event, hex_id(handler))
            new_state.targets_by_event[event] = target
            self._event_handlers[target] = handler
            model_event_handlers[event] = {
                "target": target,
                "preventDefault": handler.prevent_default,
                "stopPropagation": handler.stop_propagation,
            }

        return None

    def _render_model_event_handlers_without_old_state(
        self,
        new_state: _ModelState,
        handlers_by_event: Dict[str, EventHandler],
    ) -> None:
        if not handlers_by_event:
            return None

        model_event_handlers = new_state.model["eventHandlers"] = {}
        for event, handler in handlers_by_event.items():
            target = hex_id(handler)
            new_state.targets_by_event[event] = target
            self._event_handlers[target] = handler
            model_event_handlers[event] = {
                "target": target,
                "preventDefault": handler.prevent_default,
                "stopPropagation": handler.stop_propagation,
            }

        return None

    def _render_model_children(
        self,
        old_state: Optional[_ModelState],
        new_state: _ModelState,
        raw_children: Any,
    ) -> None:
        if not isinstance(raw_children, (list, tuple)):
            raw_children = [raw_children]

        if old_state is None:
            if raw_children:
                self._render_model_children_without_old_state(new_state, raw_children)
            return None
        elif not raw_children:
            self._unmount_model_states(list(old_state.children_by_key.values()))
            return None

        child_type_key_tuples = list(_process_child_type_and_key(raw_children))

        new_keys = {item[2] for item in child_type_key_tuples}
        if len(new_keys) != len(raw_children):
            key_counter = Counter(item[2] for item in child_type_key_tuples)
            duplicate_keys = [key for key, count in key_counter.items() if count > 1]
            raise ValueError(
                f"Duplicate keys {duplicate_keys} at {new_state.patch_path or '/'!r}"
            )

        old_keys = set(old_state.children_by_key).difference(new_keys)
        if old_keys:
            self._unmount_model_states(
                [old_state.children_by_key[key] for key in old_keys]
            )

        new_children = new_state.model["children"] = []
        for index, (child, child_type, key) in enumerate(child_type_key_tuples):
            if child_type is _DICT_TYPE:
                old_child_state = old_state.children_by_key.get(key)
                if old_child_state is not None:
                    new_child_state = old_child_state.new(new_state, None)
                else:
                    new_child_state = _ModelState(new_state, index, key, None)
                self._render_model(old_child_state, new_child_state, child)
                new_children.append(new_child_state.model)
                new_state.children_by_key[key] = new_child_state
            elif child_type is _COMPONENT_TYPE:
                old_child_state = old_state.children_by_key.get(key)
                if old_child_state is not None:
                    new_child_state = old_child_state.new(new_state, child)
                else:
                    hook = LifeCycleHook(self, child)
                    new_child_state = _ModelState(new_state, index, key, hook)
                self._render_component(old_child_state, new_child_state, child)
            else:
                new_children.append(child)

    def _render_model_children_without_old_state(
        self, new_state: _ModelState, raw_children: List[Any]
    ) -> None:
        new_children = new_state.model["children"] = []
        for index, (child, child_type, key) in enumerate(
            _process_child_type_and_key(raw_children)
        ):
            if child_type is _DICT_TYPE:
                child_state = _ModelState(new_state, index, key, None)
                self._render_model(None, child_state, child)
                new_children.append(child_state.model)
                new_state.children_by_key[key] = child_state
            elif child_type is _COMPONENT_TYPE:
                life_cycle_hook = LifeCycleHook(self, child)
                child_state = _ModelState(new_state, index, key, life_cycle_hook)
                self._render_component(None, child_state, child)
            else:
                new_children.append(child)

    def _unmount_model_states(self, old_states: List[_ModelState]) -> None:
        to_unmount = old_states[::-1]
        while to_unmount:
            state = to_unmount.pop()
            if hasattr(state, "life_cycle_hook"):
                hook = state.life_cycle_hook
                hook.component_will_unmount()
                del self._model_state_by_component_id[id(hook.component)]
            to_unmount.extend(state.children_by_key.values())

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.root})"


class _ModelState:

    __slots__ = (
        "index",
        "key",
        "_parent_ref",
        "life_cycle_hook",
        "component",
        "patch_path",
        "model",
        "targets_by_event",
        "children_by_key",
        "__weakref__",
    )

    model: _ModelVdom
    life_cycle_hook: LifeCycleHook
    patch_path: str
    component: AbstractComponent

    def __init__(
        self,
        parent: Optional[_ModelState],
        index: int,
        key: Any,
        life_cycle_hook: Optional[LifeCycleHook],
    ) -> None:
        self.index = index
        self.key = key

        if parent is not None:
            self._parent_ref = ref(parent)
            self.patch_path = f"{parent.patch_path}/children/{index}"
        else:
            self.patch_path = ""

        if life_cycle_hook is not None:
            self.life_cycle_hook = life_cycle_hook
            self.component = life_cycle_hook.component

        self.targets_by_event: Dict[str, str] = {}
        self.children_by_key: Dict[str, _ModelState] = {}

    @property
    def parent(self) -> _ModelState:
        # An AttributeError here is ok. It's synonymous
        # with the existance of 'parent' attribute
        p = self._parent_ref()
        assert p is not None, "detached model state"
        return p

    def new(
        self,
        new_parent: Optional[_ModelState],
        component: Optional[AbstractComponent],
    ) -> _ModelState:
        if new_parent is None:
            new_parent = getattr(self, "parent", None)

        life_cycle_hook: Optional[LifeCycleHook]
        if hasattr(self, "life_cycle_hook"):
            assert component is not None
            life_cycle_hook = self.life_cycle_hook
            life_cycle_hook.component = component
        else:
            life_cycle_hook = None

        return _ModelState(new_parent, self.index, self.key, life_cycle_hook)

    def iter_children(self, include_self: bool = True) -> Iterator[_ModelState]:
        to_yield = [self] if include_self else []
        while to_yield:
            node = to_yield.pop()
            yield node
            to_yield.extend(node.children_by_key.values())


class _ComponentQueue:

    __slots__ = "_loop", "_queue", "_pending"

    def __init__(self) -> None:
        self._loop = asyncio.get_event_loop()
        self._queue: "asyncio.Queue[AbstractComponent]" = asyncio.Queue()
        self._pending: Set[int] = set()

    def put(self, component: AbstractComponent) -> None:
        component_id = id(component)
        if component_id not in self._pending:
            self._pending.add(component_id)
            self._loop.call_soon_threadsafe(self._queue.put_nowait, component)
        return None

    async def get(self) -> AbstractComponent:
        component = await self._queue.get()
        self._pending.remove(id(component))
        return component


def _process_child_type_and_key(
    children: List[Any],
) -> Iterator[Tuple[Any, int, Any]]:
    for index, child in enumerate(children):
        if isinstance(child, dict):
            child_type = _DICT_TYPE
            key = child.get("key")
        elif isinstance(child, AbstractComponent):
            child_type = _COMPONENT_TYPE
            key = getattr(child, "key", None)
        else:
            child = f"{child}"
            child_type = _STRING_TYPE
            key = None

        if key is None:
            key = _default_key(index)

        yield (child, child_type, key)


if IDOM_FEATURE_INDEX_AS_DEFAULT_KEY.current:

    def _default_key(index: int) -> Any:  # pragma: no cover
        return index


else:

    def _default_key(index: int) -> Any:
        return object()


# used in _process_child_type_and_key
_DICT_TYPE = 1
_COMPONENT_TYPE = 2
_STRING_TYPE = 3


class _ModelEventTarget(TypedDict):
    target: str
    preventDefault: bool  # noqa
    stopPropagation: bool  # noqa


class _ModelImportSource(TypedDict):
    source: str
    fallback: Any


class _ModelVdomOptional(TypedDict, total=False):
    key: str  # noqa
    children: List[Any]  # noqa
    attributes: Dict[str, Any]  # noqa
    eventHandlers: Dict[str, _ModelEventTarget]  # noqa
    importSource: _ModelImportSource  # noqa


class _ModelVdomRequired(TypedDict, total=True):
    tagName: str  # noqa


class _ModelVdom(_ModelVdomRequired, _ModelVdomOptional):
    """A VDOM dictionary model specifically for use with a :class:`Layout`"""



from __future__ import print_function

import pipes
import shutil
import subprocess
import sys
import traceback
from distutils import log
from distutils.command.build import build  # type: ignore
from distutils.command.sdist import sdist  # type: ignore
from pathlib import Path

from setuptools import find_packages, setup
from setuptools.command.develop import develop


if sys.platform == "win32":
    from subprocess import list2cmdline
else:

    def list2cmdline(cmd_list):
        return " ".join(map(pipes.quote, cmd_list))


# the name of the project
name = "idom"

# basic paths used to gather files
root_dir = Path(__file__).parent
src_dir = root_dir / "src"
package_dir = src_dir / name


# -----------------------------------------------------------------------------
# Package Definition
# -----------------------------------------------------------------------------


package = {
    "name": name,
    "python_requires": ">=3.7",
    "packages": find_packages(str(src_dir)),
    "package_dir": {"": "src"},
    "description": "Control the web with Python",
    "author": "Ryan Morshead",
    "author_email": "ryan.morshead@gmail.com",
    "url": "https://github.com/rmorshea/idom",
    "license": "MIT",
    "platforms": "Linux, Mac OS X, Windows",
    "keywords": ["interactive", "widgets", "DOM", "React"],
    "include_package_data": True,
    "zip_safe": False,
    "setup_requires": ["setuptools_scm"],
    "use_scm_version": True,
    "classifiers": [
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Multimedia :: Graphics",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
}


# -----------------------------------------------------------------------------
# CLI Entrypoints
# -----------------------------------------------------------------------------


package["entry_points"] = {
    "console_scripts": ["idom = idom.__main__:main"],
}


# -----------------------------------------------------------------------------
# Requirements
# -----------------------------------------------------------------------------


requirements = []
with (root_dir / "requirements" / "pkg-deps.txt").open() as f:
    for line in map(str.strip, f):
        if not line.startswith("#"):
            requirements.append(line)
package["install_requires"] = requirements

_current_extras = []
extra_requirements = {"all": []}  # type: ignore
extra_requirements_path = root_dir / "requirements" / "pkg-extras.txt"
with extra_requirements_path.open() as f:
    for line in map(str.strip, f):
        if line.startswith("#") and line[1:].strip().startswith("extra="):
            _current_extras = [e.strip() for e in line.split("=", 1)[1].split(",")]
            if "all" in _current_extras:
                raise ValueError("%r uses the reserved extra name 'all'")
            for e in _current_extras:
                extra_requirements[e] = []
        elif _current_extras:
            for e in _current_extras:
                extra_requirements[e].append(line)
            extra_requirements["all"].append(line)
        elif line:
            raise ValueError(
                f"No '# extra=<name>' header before requirements in {extra_requirements_path}"
            )
package["extras_require"] = extra_requirements


# -----------------------------------------------------------------------------
# Library Description
# -----------------------------------------------------------------------------


with (root_dir / "README.md").open() as f:
    long_description = f.read()

package["long_description"] = long_description
package["long_description_content_type"] = "text/markdown"


# ----------------------------------------------------------------------------
# Build Javascript
# ----------------------------------------------------------------------------


def build_javascript_first(cls):
    class Command(cls):
        def run(self):
            log.info("Installing Javascript...")
            try:
                js_dir = str(package_dir / "client" / "app")
                npm = shutil.which("npm")  # this is required on windows
                if npm is None:
                    raise RuntimeError("NPM is not installed.")
                for args in (f"{npm} install", f"{npm} run build"):
                    args_list = args.split()
                    log.info(f"> {list2cmdline(args_list)}")
                    subprocess.run(args_list, cwd=js_dir, check=True)
            except Exception:
                log.error("Failed to install Javascript")
                log.error(traceback.format_exc())
                raise
            else:
                log.info("Successfully installed Javascript")
            super().run()

    return Command


package["cmdclass"] = {
    "sdist": build_javascript_first(sdist),
    "build": build_javascript_first(build),
    "develop": build_javascript_first(develop),
}


# -----------------------------------------------------------------------------
# Install It
# -----------------------------------------------------------------------------


if __name__ == "__main__":
    setup(**package)



"""
Events
"""

import asyncio
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Union,
)

from anyio import create_task_group


EventsMapping = Union[Dict[str, Union["Callable[..., Any]", "EventHandler"]], "Events"]


def event(
    function: Optional[Callable[..., Any]] = None,
    stop_propagation: bool = False,
    prevent_default: bool = False,
) -> Union["EventHandler", Callable[[Callable[..., Any]], "EventHandler"]]:
    """Create an event handler function with extra functionality.

    While you're always free to add callbacks by assigning them to an element's attributes

    .. code-block:: python

        element = idom.html.button({"onClick": my_callback})

    You may want the ability to prevent the default action associated with the event
    from taking place, or stoping the event from propagating up the DOM. This decorator
    allows you to add that functionality to your callbacks.

    Parameters:
        function:
            A callback responsible for handling the event.
        stop_propagation:
            Block the event from propagating further up the DOM.
        prevent_default:
            Stops the default actional associate with the event from taking place.
    """
    handler = EventHandler(stop_propagation, prevent_default)
    if function is not None:
        handler.add(function)
        return handler
    else:
        return handler.add


class Events(Mapping[str, "EventHandler"]):
    """A container for event handlers.

    Assign this object to the ``"eventHandlers"`` field of an element model.
    """

    __slots__ = "_handlers"

    def __init__(self) -> None:
        self._handlers: Dict[str, EventHandler] = {}

    def on(
        self, event: str, stop_propagation: bool = False, prevent_default: bool = False
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """A decorator for adding an event handler.

        Parameters:
            event:
                The camel-case name of the event, the word "on" is automatically
                prepended. So passing "keyDown" would refer to the event "onKeyDown".
            stop_propagation:
                Block the event from propagating further up the DOM.
            prevent_default:
                Stops the default action associate with the event from taking place.

        Returns:
            A decorator which accepts an event handler function as its first argument.
            The parameters of the event handler function may indicate event attributes
            which should be sent back from the frontend. See :class:`EventHandler` for
            more info.

        Examples:
            Simple "onClick" event handler:

            .. code-block:: python

                def clickable_element():
                    events = Events()

                    @events.on("click")
                    def handler(event):
                        # do something on a click event
                        ...

                    return idom.vdom("button", "hello!", eventHandlers=events)
        """
        if not event.startswith("on"):
            event = "on" + event[:1].upper() + event[1:]

        if event not in self._handlers:
            handler = EventHandler(stop_propagation, prevent_default)
            self._handlers[event] = handler
        else:
            handler = self._handlers[event]

        def setup(function: Callable[..., Any]) -> Callable[..., Any]:
            handler.add(function)
            return function

        return setup

    def __contains__(self, key: Any) -> bool:
        return key in self._handlers

    def __len__(self) -> int:
        return len(self._handlers)

    def __iter__(self) -> Iterator[str]:
        return iter(self._handlers)

    def __getitem__(self, key: str) -> "EventHandler":
        return self._handlers[key]

    def __repr__(self) -> str:  # pragma: no cover
        return repr(self._handlers)


class EventHandler:
    """An object which defines an event handler.

    The event handler object acts like a coroutine when called.

    Parameters:
        stop_propagation:
            Block the event from propagating further up the DOM.
        prevent_default:
            Stops the default action associate with the event from taking place.
    """

    __slots__ = (
        "__weakref__",
        "_coro_handlers",
        "_func_handlers",
        "prevent_default",
        "stop_propagation",
    )

    def __init__(
        self,
        stop_propagation: bool = False,
        prevent_default: bool = False,
    ) -> None:
        self.stop_propagation = stop_propagation
        self.prevent_default = prevent_default
        self._coro_handlers: List[Callable[..., Coroutine[Any, Any, Any]]] = []
        self._func_handlers: List[Callable[..., Any]] = []

    def add(self, function: Callable[..., Any]) -> "EventHandler":
        """Add a callback function or coroutine to the event handler.

        Parameters:
            function:
                The event handler function accepting parameters sent by the client.
                Typically this is a single ``event`` parameter that is a dictionary.
        """
        if asyncio.iscoroutinefunction(function):
            self._coro_handlers.append(function)
        else:
            self._func_handlers.append(function)
        return self

    def remove(self, function: Callable[..., Any]) -> None:
        """Remove the given function or coroutine from this event handler.

        Raises:
            ValueError: if not found
        """
        if asyncio.iscoroutinefunction(function):
            self._coro_handlers.remove(function)
        else:
            self._func_handlers.remove(function)

    def clear(self) -> None:
        """Remove all functions and coroutines from this event handler"""
        self._coro_handlers.clear()
        self._func_handlers.clear()

    async def __call__(self, data: List[Any]) -> Any:
        """Trigger all callbacks in the event handler."""
        if self._coro_handlers:
            async with create_task_group() as group:
                for handler in self._coro_handlers:
                    group.start_soon(handler, *data)
        for handler in self._func_handlers:
            handler(*data)

    def __contains__(self, function: Any) -> bool:
        if asyncio.iscoroutinefunction(function):
            return function in self._coro_handlers
        else:
            return function in self._func_handlers



"""
Component
=========
"""

from __future__ import annotations

import abc
import inspect
from functools import wraps
from typing import Any, Callable, Dict, Optional, Tuple, Union

from .utils import hex_id
from .vdom import VdomDict


ComponentConstructor = Callable[..., "AbstractComponent"]
ComponentRenderFunction = Callable[..., Union["AbstractComponent", VdomDict]]


def component(function: ComponentRenderFunction) -> Callable[..., "Component"]:
    """A decorator for defining an :class:`Component`.

    Parameters:
        function: The function that will render a :class:`VdomDict`.
    """

    @wraps(function)
    def constructor(*args: Any, key: Optional[Any] = None, **kwargs: Any) -> Component:
        return Component(function, key, args, kwargs)

    return constructor


class AbstractComponent(abc.ABC):

    __slots__ = ["key"]
    if not hasattr(abc.ABC, "__weakref__"):
        __slots__.append("__weakref__")  # pragma: no cover

    key: Optional[Any]

    @abc.abstractmethod
    def render(self) -> VdomDict:
        """Render the component's :class:`VdomDict`."""


class Component(AbstractComponent):
    """An object for rending component models."""

    __slots__ = "_func", "_args", "_kwargs"

    def __init__(
        self,
        function: ComponentRenderFunction,
        key: Optional[Any],
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> None:
        self.key = key
        self._func = function
        self._args = args
        self._kwargs = kwargs
        if key is not None:
            kwargs["key"] = key

    def render(self) -> VdomDict:
        model = self._func(*self._args, **self._kwargs)
        if isinstance(model, AbstractComponent):
            model = {"tagName": "div", "children": [model]}
        return model

    def __repr__(self) -> str:
        sig = inspect.signature(self._func)
        try:
            args = sig.bind(*self._args, **self._kwargs).arguments
        except TypeError:
            return f"{self._func.__name__}(...)"
        else:
            items = ", ".join(f"{k}={v!r}" for k, v in args.items())
            if items:
                return f"{self._func.__name__}({hex_id(self)}, {items})"
            else:
                return f"{self._func.__name__}({hex_id(self)})"



"""
VDOM Constructors
=================
"""

from __future__ import annotations

import logging
from typing import Any, Iterable, List, Mapping, Optional, Sequence, Tuple, Union

from fastjsonschema import compile as compile_json_schema
from mypy_extensions import TypedDict
from typing_extensions import Protocol

from idom.config import IDOM_DEBUG_MODE

from .events import EventHandler


logger = logging.getLogger()


VDOM_JSON_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "$ref": "#/definitions/element",
    "definitions": {
        "element": {
            "type": "object",
            "properties": {
                "tagName": {"type": "string"},
                "key": {"type": "string"},
                "children": {"$ref": "#/definitions/elementChildren"},
                "attributes": {"type": "object"},
                "eventHandlers": {"$ref": "#/definitions/elementEventHandlers"},
                "importSource": {"$ref": "#/definitions/importSource"},
            },
            "required": ["tagName"],
        },
        "elementChildren": {
            "type": "array",
            "items": {"$ref": "#/definitions/elementOrString"},
        },
        "elementEventHandlers": {
            "type": "object",
            "patternProperties": {
                ".*": {"$ref": "#/definitions/eventHander"},
            },
        },
        "eventHander": {
            "type": "object",
            "properties": {
                "target": {"type": "string"},
                "preventDefault": {"type": "boolean"},
                "stopPropagation": {"type": "boolean"},
            },
            "required": ["target"],
        },
        "importSource": {
            "type": "object",
            "properties": {
                "source": {"type": "string"},
                "sourceType": {"enum": ["URL", "NAME"]},
                "fallback": {
                    "type": ["object", "string", "null"],
                    "if": {"not": {"type": "null"}},
                    "then": {"$ref": "#/definitions/elementOrString"},
                },
            },
            "required": ["source"],
        },
        "elementOrString": {
            "type": ["object", "string"],
            "if": {"type": "object"},
            "then": {"$ref": "#/definitions/element"},
        },
    },
}
"""JSON Schema describing serialized VDOM - see :ref:`VDOM Mimetype` for more info"""


# we can't add a docstring to this because Sphinx doesn't know how to find its source
_COMPILED_VDOM_VALIDATOR = compile_json_schema(VDOM_JSON_SCHEMA)


def validate_vdom(value: Any) -> None:
    """Validate serialized VDOM - see :attr:`VDOM_JSON_SCHEMA` for more info"""
    _COMPILED_VDOM_VALIDATOR(value)


class ImportSourceDict(TypedDict):
    source: str
    fallback: Any
    sourceType: str  # noqa


class _VdomDictOptional(TypedDict, total=False):
    key: str  # noqa
    children: Sequence[Any]  # noqa
    attributes: Mapping[str, Any]  # noqa
    eventHandlers: Mapping[str, EventHandler]  # noqa
    importSource: ImportSourceDict  # noqa


class _VdomDictRequired(TypedDict, total=True):
    tagName: str  # noqa


class VdomDict(_VdomDictRequired, _VdomDictOptional):
    """A VDOM dictionary"""


_AttributesAndChildrenArg = Union[Mapping[str, Any], str, Iterable[Any], Any]
_EventHandlersArg = Optional[Mapping[str, EventHandler]]
_ImportSourceArg = Optional[ImportSourceDict]


def vdom(
    tag: str,
    *attributes_and_children: _AttributesAndChildrenArg,
    key: str = "",
    event_handlers: _EventHandlersArg = None,
    import_source: _ImportSourceArg = None,
) -> VdomDict:
    """A helper function for creating VDOM dictionaries.

    Parameters:
        tag:
            The type of element (e.g. 'div', 'h1', 'img')
        attributes_and_children:
            An optional attribute mapping followed by any number of children or
            iterables of children. The attribute mapping **must** precede the children,
            or children which will be merged into their respective parts of the model.
        key:
            A string idicating the identity of a particular element. This is significant
            to preserve event handlers across updates - without a key, a re-render would
            cause these handlers to be deleted, but with a key, they would be redirected
            to any newly defined handlers.
        event_handlers:
            Maps event types to coroutines that are responsible for handling those events.
        import_source:
            (subject to change) specifies javascript that, when evaluated returns a
            React component.
    """
    model: VdomDict = {"tagName": tag}

    attributes, children = coalesce_attributes_and_children(attributes_and_children)

    if attributes:
        model["attributes"] = attributes

    if children:
        model["children"] = children

    if key:
        model["key"] = key

    if event_handlers is not None:
        model["eventHandlers"] = event_handlers

    if import_source is not None:
        model["importSource"] = import_source

    return model


class VdomDictConstructor(Protocol):
    def __call__(
        self,
        *args: _AttributesAndChildrenArg,
        event_handlers: _EventHandlersArg = None,
        import_source: _ImportSourceArg = None,
    ) -> VdomDict:
        ...


def make_vdom_constructor(tag: str, allow_children: bool = True) -> VdomDictConstructor:
    """Return a constructor for VDOM dictionaries with the given tag name.

    The resulting callable will have the same interface as :func:`vdom` but without its
    first ``tag`` argument.
    """

    def constructor(
        *attributes_and_children: _AttributesAndChildrenArg,
        key: str = "",
        event_handlers: _EventHandlersArg = None,
        import_source: _ImportSourceArg = None,
    ) -> VdomDict:
        model = vdom(
            tag,
            *attributes_and_children,
            key=key,
            event_handlers=event_handlers,
            import_source=import_source,
        )
        if not allow_children and "children" in model:
            raise TypeError(f"{tag!r} nodes cannot have children.")
        return model

    constructor.__name__ = tag
    qualname_prefix = constructor.__qualname__.rsplit(".", 1)[0]
    constructor.__qualname__ = qualname_prefix + f".{tag}"
    constructor.__doc__ = (
        f"""Create a new ``<{tag}/>`` - returns :ref:`VDOM <VDOM Mimetype>`."""
    )

    return constructor


def coalesce_attributes_and_children(
    values: Sequence[Any],
) -> Tuple[Mapping[str, Any], List[Any]]:
    if not values:
        return {}, []

    children_or_iterables: Sequence[Any]
    attributes, *children_or_iterables = values
    if not _is_attributes(attributes):
        attributes = {}
        children_or_iterables = values

    children: List[Any] = []
    for child in children_or_iterables:
        if _is_single_child(child):
            children.append(child)
        else:
            children.extend(child)

    return attributes, children


def _is_attributes(value: Any) -> bool:
    return isinstance(value, Mapping) and "tagName" not in value


if IDOM_DEBUG_MODE.current:

    _debug_is_attributes = _is_attributes

    def _is_attributes(value: Any) -> bool:
        result = _debug_is_attributes(value)
        if result and "children" in value:
            logger.error(f"Reserved key 'children' found in attributes {value}")
        return result


def _is_single_child(value: Any) -> bool:
    return isinstance(value, (str, Mapping)) or not hasattr(value, "__iter__")


if IDOM_DEBUG_MODE.current:

    _debug_is_single_child = _is_single_child

    def _is_single_child(value: Any) -> bool:
        if _debug_is_single_child(value):
            return True

        from .component import AbstractComponent

        if hasattr(value, "__iter__") and not hasattr(value, "__len__"):
            logger.error(
                f"Did not verify key-path integrity of children in generator {value} "
                "- pass a sequence (i.e. list of finite length) in order to verify"
            )
        else:
            for child in value:
                if (isinstance(child, AbstractComponent) and child.key is None) or (
                    isinstance(child, Mapping) and "key" not in child
                ):
                    logger.error(f"Key not specified for dynamic child {child}")

        return False



"""
Client Modules
==============
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union, overload
from urllib.parse import urlparse

from idom.core.vdom import ImportSourceDict, VdomDict, make_vdom_constructor

from . import _private, manage


@overload
def install(
    packages: str,
    ignore_installed: bool,
    fallback: Optional[str],
) -> Module:
    ...


@overload
def install(
    packages: Union[List[str], Tuple[str]],
    ignore_installed: bool,
    fallback: Optional[str],
) -> List[Module]:
    ...


def install(
    packages: Union[str, List[str], Tuple[str]],
    ignore_installed: bool = False,
    fallback: Optional[str] = None,
) -> Union[Module, List[Module]]:
    return_one = False
    if isinstance(packages, str):
        packages = [packages]
        return_one = True

    pkg_names = [_private.get_package_name(pkg) for pkg in packages]

    if ignore_installed or set(pkg_names).difference(manage.web_module_names()):
        manage.build(packages, clean_build=False)

    if return_one:
        return Module(pkg_names[0], fallback=fallback)
    else:
        return [Module(pkg, fallback=fallback) for pkg in pkg_names]


NAME_SOURCE = "NAME"
"""A named souce - usually a Javascript package name"""

URL_SOURCE = "URL"
"""A source loaded from a URL, usually from a CDN"""

SOURCE_TYPES = {NAME_SOURCE, URL_SOURCE}
"""The possible source types for a :class:`Module`"""


class Module:
    """A Javascript module

    Parameters:
        source:
            The URL to an ECMAScript module which exports React components
            (*with* a ``.js`` file extension) or name of a module installed in the
            built-in client application (*without* a ``.js`` file extension).
        source_type:
            The type of the given ``source``. See :const:`SOURCE_TYPES` for the set of
            possible values.
        file:
            Only applicable if running on a client app which supports this feature.
            Dynamically install the code in the give file as a single-file module. The
            built-in client will inject this module adjacent to other installed modules
            which means they can be imported via a relative path like
            ``./some-other-installed-module.js``.
        fallack:
            What to display while the modules is being loaded.

    Attributes:
        installed:
            Whether or not this module has been installed into the built-in client app.
        url:
            The URL this module will be imported from.
    """

    __slots__ = "source", "source_type", "fallback", "exports"

    def __init__(
        self,
        source: str,
        source_type: Optional[str] = None,
        source_file: Optional[Union[str, Path]] = None,
        fallback: Optional[str] = None,
        check_exports: Optional[bool] = None,
    ) -> None:
        self.source = source
        self.fallback = fallback
        self.exports: Optional[Set[str]] = None

        if source_type is None:
            self.source_type = URL_SOURCE if _is_url(source) else NAME_SOURCE
        elif source_type in SOURCE_TYPES:
            self.source_type = source_type
        else:
            raise ValueError(f"Invalid source type {source_type!r}")

        if self.source_type == URL_SOURCE:
            if check_exports is True:
                raise ValueError(f"Can't check exports for source type {source_type!r}")
            elif source_file is not None:
                raise ValueError(f"File given, but source type is {source_type!r}")
            else:
                return None
        elif check_exports is None:
            check_exports = True

        if source_file is not None:
            manage.add_web_module(source, source_file)
        elif not manage.web_module_exists(source):
            raise ValueError(f"Module {source!r} does not exist")

        if check_exports:
            self.exports = manage.web_module_exports(source)

    def declare(
        self,
        name: str,
        has_children: bool = True,
        fallback: Optional[str] = None,
    ) -> Import:
        """Return  an :class:`Import` for the given :class:`Module` and ``name``

        This roughly translates to the javascript statement

        .. code-block:: javascript

            import { name } from "module"

        Where ``name`` is the given name, and ``module`` is the :attr:`Module.url` of
        this :class:`Module` instance.
        """
        if self.exports is not None and name not in self.exports:
            raise ValueError(
                f"{self} does not export {name!r}, available options are {list(self.exports)}"
            )

        return Import(
            name,
            self.source,
            self.source_type,
            has_children,
            fallback or self.fallback,
        )

    def __getattr__(self, name: str) -> Import:
        if name[0].lower() == name[0]:
            # component names should be capitalized
            raise AttributeError(f"{self} has no attribute {name!r}")
        return self.declare(name)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Module)
            and self.source == other.source
            and self.source_type == other.source_type
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.source})"


class Import:
    """Import a react module

    Once imported, you can instantiate the library's components by calling them
    via attribute-access.

    Examples:

        .. code-block:: python

            victory = idom.Import("victory", "VictoryBar" install=True)
            style = {"parent": {"width": "500px"}}
            victory.VictoryBar({"style": style}, fallback="loading...")
    """

    __slots__ = "_constructor", "_import_source", "_name"

    def __init__(
        self,
        name: str,
        source: str,
        source_type: str,
        has_children: bool = True,
        fallback: Optional[str] = None,
    ) -> None:
        self._name = name
        self._constructor = make_vdom_constructor(name, has_children)
        self._import_source = ImportSourceDict(
            source=source,
            sourceType=source_type,
            fallback=fallback,
        )

    def __call__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> VdomDict:
        return self._constructor(import_source=self._import_source, *args, **kwargs)

    def __repr__(self) -> str:
        info: Dict[str, Any] = {"name": self._name, **self._import_source}
        strings = ", ".join(f"{k}={v!r}" for k, v in info.items())
        return f"{type(self).__name__}({strings})"


def _is_url(string: str) -> bool:
    if string.startswith("/") or string.startswith("./") or string.startswith("../"):
        return True
    else:
        parsed = urlparse(string)
        return bool(parsed.scheme and parsed.netloc)
